import json

import click
from globus_sdk.base import safe_stringify

from globus_cli.safeio.get_option_vals import outformat_is_json


class PrintableErrorField(object):
    """
    A glorified tuple with a kwarg in its constructor.
    Coerces name and value fields to unicode for output consistency
    """

    TEXT_PREFIX = "Globus CLI Error:"

    def __init__(self, name, value, multiline=False):
        self.multiline = multiline
        self.name = safe_stringify(name)
        self.raw_value = safe_stringify(value)
        self.value = self._format_value(self.raw_value)

    @property
    def _text_prefix_len(self):
        return len(self.TEXT_PREFIX)

    def _format_value(self, val):
        """
        formats a value to be good for textmode printing
        val must be unicode
        """
        name = self.name + ":"
        if not self.multiline or "\n" not in val:
            val = u"{0} {1}".format(name.ljust(self._text_prefix_len), val)
        else:
            spacer = "\n" + " " * (self._text_prefix_len + 1)
            val = u"{0}{1}{2}".format(name, spacer, spacer.join(val.split("\n")))

        return val


def write_error_info(error_name, fields, message=None):

    if outformat_is_json():
        # dictify joined tuple lists and dump to json string
        message = click.style(
            json.dumps(
                dict(
                    [("error_name", error_name)]
                    + [(f.name, f.raw_value) for f in fields]
                ),
                indent=2,
                separators=(",", ": "),
                sort_keys=True,
            ),
            fg="yellow",
        )
    if not message:
        message = u"A{0} {1} Occurred.\n{2}".format(
            "n" if error_name[0] in "aeiouAEIOU" else "",
            click.style(error_name, bold=True, fg="red"),
            click.style("\n".join(f.value for f in fields), fg="yellow"),
        )
        message = u"{0} {1}".format(PrintableErrorField.TEXT_PREFIX, message)

    click.echo(message, err=True)



__all__ = ["EXPLICIT_NULL"]


# this object is a sentinel value used to disambiguate values which are being
# intentionally nulled from values which are incidentally `None` because no
# argument was provided
EXPLICIT_NULL = object()



from globus_cli.parsing import (
    command,
    endpoint_id_arg,
    server_add_and_update_opts,
    server_id_arg,
)
from globus_cli.safeio import FORMAT_TEXT_RAW, formatted_print
from globus_cli.services.transfer import assemble_generic_doc, get_client


@command("update")
@server_add_and_update_opts
@endpoint_id_arg
@server_id_arg
def server_update(
    endpoint_id,
    server_id,
    subject,
    port,
    scheme,
    hostname,
    incoming_data_ports,
    outgoing_data_ports,
):
    """Update attributes of a server on an endpoint"""
    client = get_client()

    server_doc = assemble_generic_doc(
        "server", subject=subject, port=port, scheme=scheme, hostname=hostname
    )

    # n.b. must be done after assemble_generic_doc(), as that function filters
    # out `None`s, which we need to be able to set for `'unspecified'`
    if incoming_data_ports:
        server_doc.update(
            incoming_data_port_start=incoming_data_ports[0],
            incoming_data_port_end=incoming_data_ports[1],
        )
    if outgoing_data_ports:
        server_doc.update(
            outgoing_data_port_start=outgoing_data_ports[0],
            outgoing_data_port_end=outgoing_data_ports[1],
        )

    res = client.update_endpoint_server(endpoint_id, server_id, server_doc)
    formatted_print(res, text_format=FORMAT_TEXT_RAW, response_key="message")



from globus_cli.parsing import command, endpoint_id_arg
from globus_cli.safeio import FORMAT_TEXT_RAW, formatted_print
from globus_cli.services.transfer import get_client


@command("deactivate")
@endpoint_id_arg
def endpoint_deactivate(endpoint_id):
    """Deactivate an endpoint"""
    client = get_client()
    res = client.endpoint_deactivate(endpoint_id)
    formatted_print(res, text_format=FORMAT_TEXT_RAW, response_key="message")



import click

from globus_cli.config import get_config_obj
from globus_cli.parsing import command


@command("set", disable_options=["format"])
@click.argument("parameter", required=True)
@click.argument("value", required=True)
def set_command(value, parameter):
    """Set a value in the Globus config file"""
    conf = get_config_obj()

    section = "cli"
    if "." in parameter:
        section, parameter = parameter.split(".", 1)

    # ensure that the section exists
    if section not in conf:
        conf[section] = {}
    # set the value for the given parameter
    conf[section][parameter] = value

    # write to disk
    click.echo("Writing updated config to {}".format(conf.filename))
    conf.write()



import click

from globus_cli.commands.bookmark.helpers import resolve_id_or_name
from globus_cli.parsing import command
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print, is_verbose
from globus_cli.services.transfer import get_client


@command("show")
@click.argument("bookmark_id_or_name")
def bookmark_show(bookmark_id_or_name):
    """
    Given a bookmark name or ID resolves bookmark in endpoint:path format.
    Use --verbose for additional fields.
    """
    client = get_client()
    res = resolve_id_or_name(client, bookmark_id_or_name)
    formatted_print(
        res,
        text_format=FORMAT_TEXT_RECORD,
        fields=(
            ("ID", "id"),
            ("Name", "name"),
            ("Endpoint ID", "endpoint_id"),
            ("Path", "path"),
        ),
        simple_text=(
            # standard output is endpoint:path format
            "{}:{}".format(res["endpoint_id"], res["path"])
            # verbose output includes all fields
            if not is_verbose()
            else None
        ),
    )



from globus_cli.parsing import command, endpoint_id_arg, role_id_arg
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print
from globus_cli.services.auth import lookup_identity_name
from globus_cli.services.transfer import get_client


def lookup_principal(role):
    return lookup_identity_name(role["principal"])


@command("show")
@endpoint_id_arg
@role_id_arg
def role_show(endpoint_id, role_id):
    """Show full info for a role on an endpoint"""
    client = get_client()

    role = client.get_endpoint_role(endpoint_id, role_id)
    formatted_print(
        role,
        text_format=FORMAT_TEXT_RECORD,
        fields=(
            ("Principal Type", "principal_type"),
            ("Principal", lookup_principal),
            ("Role", "role"),
        ),
    )



import webbrowser

import click

from globus_cli.config import (
    AUTH_AT_EXPIRES_OPTNAME,
    AUTH_AT_OPTNAME,
    AUTH_RT_OPTNAME,
    TRANSFER_AT_EXPIRES_OPTNAME,
    TRANSFER_AT_OPTNAME,
    TRANSFER_RT_OPTNAME,
    internal_auth_client,
    lookup_option,
    write_option,
)
from globus_cli.helpers.local_server import LocalServerError, start_local_server

SCOPES = (
    "openid profile email "
    "urn:globus:auth:scope:auth.globus.org:view_identity_set "
    "urn:globus:auth:scope:transfer.api.globus.org:all"
)


def do_link_auth_flow(session_params=None, force_new_client=False):
    """
    Prompts the user with a link to authenticate with globus auth
    and authorize the CLI to act on their behalf.
    """
    session_params = session_params or {}

    # get the ConfidentialApp client object
    auth_client = internal_auth_client(
        requires_instance=True, force_new_client=force_new_client
    )

    # start the Confidential App Grant flow
    auth_client.oauth2_start_flow(
        redirect_uri=auth_client.base_url + "v2/web/auth-code",
        refresh_tokens=True,
        requested_scopes=SCOPES,
    )

    # prompt
    additional_params = {"prompt": "login"}
    additional_params.update(session_params)
    linkprompt = "Please authenticate with Globus here"
    click.echo(
        "{0}:\n{1}\n{2}\n{1}\n".format(
            linkprompt,
            "-" * len(linkprompt),
            auth_client.oauth2_get_authorize_url(additional_params=additional_params),
        )
    )

    # come back with auth code
    auth_code = click.prompt("Enter the resulting Authorization Code here").strip()

    # finish auth flow
    exchange_code_and_store_config(auth_client, auth_code)
    return True


def do_local_server_auth_flow(session_params=None, force_new_client=False):
    """
    Starts a local http server, opens a browser to have the user authenticate,
    and gets the code redirected to the server (no copy and pasting required)
    """
    session_params = session_params or {}

    # start local server and create matching redirect_uri
    with start_local_server(listen=("127.0.0.1", 0)) as server:
        _, port = server.socket.getsockname()
        redirect_uri = "http://localhost:{}".format(port)

        # get the ConfidentialApp client object and start a flow
        auth_client = internal_auth_client(
            requires_instance=True, force_new_client=force_new_client
        )
        auth_client.oauth2_start_flow(
            refresh_tokens=True, redirect_uri=redirect_uri, requested_scopes=SCOPES
        )
        additional_params = {"prompt": "login"}
        additional_params.update(session_params)
        url = auth_client.oauth2_get_authorize_url(additional_params=additional_params)

        # open web-browser for user to log in, get auth code
        webbrowser.open(url, new=1)
        auth_code = server.wait_for_code()

    if isinstance(auth_code, LocalServerError):
        click.echo("Authorization failed: {}".format(auth_code), err=True)
        click.get_current_context().exit(1)
    elif isinstance(auth_code, Exception):
        click.echo(
            "Authorization failed with unexpected error:\n{}".format(auth_code),
            err=True,
        )
        click.get_current_context().exit(1)

    # finish auth flow and return true
    exchange_code_and_store_config(auth_client, auth_code)
    return True


def exchange_code_and_store_config(auth_client, auth_code):
    """
    Finishes auth flow after code is gotten from command line or local server.
    Exchanges code for tokens and gets user info from auth.
    Stores tokens and user info in config.
    """
    # do a token exchange with the given code
    tkn = auth_client.oauth2_exchange_code_for_tokens(auth_code)
    tkn = tkn.by_resource_server

    # extract access tokens from final response
    transfer_at = tkn["transfer.api.globus.org"]["access_token"]
    transfer_at_expires = tkn["transfer.api.globus.org"]["expires_at_seconds"]
    transfer_rt = tkn["transfer.api.globus.org"]["refresh_token"]
    auth_at = tkn["auth.globus.org"]["access_token"]
    auth_at_expires = tkn["auth.globus.org"]["expires_at_seconds"]
    auth_rt = tkn["auth.globus.org"]["refresh_token"]

    # revoke any existing tokens
    for token_opt in (
        TRANSFER_RT_OPTNAME,
        TRANSFER_AT_OPTNAME,
        AUTH_RT_OPTNAME,
        AUTH_AT_OPTNAME,
    ):
        token = lookup_option(token_opt)
        if token:
            auth_client.oauth2_revoke_token(token)

    # write new tokens to config
    write_option(TRANSFER_RT_OPTNAME, transfer_rt)
    write_option(TRANSFER_AT_OPTNAME, transfer_at)
    write_option(TRANSFER_AT_EXPIRES_OPTNAME, transfer_at_expires)
    write_option(AUTH_RT_OPTNAME, auth_rt)
    write_option(AUTH_AT_OPTNAME, auth_at)
    write_option(AUTH_AT_EXPIRES_OPTNAME, auth_at_expires)



from __future__ import unicode_literals

import json

import click
import six
from globus_sdk import GlobusResponse

from globus_cli.safeio.awscli_text import unix_formatted_print
from globus_cli.safeio.get_option_vals import (
    get_jmespath_expression,
    outformat_is_json,
    outformat_is_unix,
)

# make sure this is a tuple
# if it's a list, pylint will freak out
__all__ = (
    "formatted_print",
    "FORMAT_SILENT",
    "FORMAT_JSON",
    "FORMAT_TEXT_TABLE",
    "FORMAT_TEXT_RECORD",
    "FORMAT_TEXT_RAW"
    # NOTE: we don't export FORMAT_TEXT_CUSTOM as actually passing it is
    # incorrect usage -- it's used internally, but similarly to the other
    # format constants
)


FORMAT_SILENT = "silent"
FORMAT_JSON = "json"
FORMAT_TEXT_TABLE = "text_table"
FORMAT_TEXT_RECORD = "text_record"
FORMAT_TEXT_RAW = "text_raw"
FORMAT_TEXT_CUSTOM = "text_custom"


def _key_to_keyfunc(k):
    """
    We allow for 'keys' which are functions that map columns onto value
    types -- they may do formatting or inspect multiple values on the
    object. In order to support this, wrap string keys in a simple function
    that does the natural lookup operation, but return any functions we
    receive as they are.
    """
    # if the key is a string, then the "keyfunc" is just a basic lookup
    # operation -- return that
    if isinstance(k, six.string_types):

        def lookup(x):
            return x[k]

        return lookup
    # otherwise, the key must be a function which is executed on the item
    # to produce a value -- return it verbatim
    return k


def _jmespath_preprocess(res):
    jmespath_expr = get_jmespath_expression()

    if isinstance(res, GlobusResponse):
        res = res.data

    if not isinstance(res, six.string_types):
        if jmespath_expr is not None:
            res = jmespath_expr.search(res)

    return res


def print_json_response(res):
    res = _jmespath_preprocess(res)
    res = json.dumps(res, indent=2, separators=(",", ": "), sort_keys=True)
    click.echo(res)


def print_unix_response(res):
    res = _jmespath_preprocess(res)
    try:
        unix_formatted_print(res)
    # Attr errors indicate that we got data which cannot be unix formatted
    # likely a scalar + non-scalar in an array, though there may be other cases
    # print good error and exit(2) (Count this as UsageError!)
    except AttributeError:
        click.echo(
            "UNIX formatting of output failed."
            "\n  "
            "This usually means that data has a structure which cannot be "
            "handled by the UNIX formatter."
            "\n  "
            "To avoid this error in the future, ensure that you query the "
            'exact properties you want from output data with "--jmespath"',
            err=True,
        )
        click.get_current_context().exit(2)


def colon_formatted_print(data, named_fields):
    maxlen = max(len(n) for n, f in named_fields) + 1
    for name, field in named_fields:
        field_keyfunc = _key_to_keyfunc(field)
        click.echo("{} {}".format((name + ":").ljust(maxlen), field_keyfunc(data)))


def print_table(iterable, headers_and_keys, print_headers=True):
    # the iterable may not be safe to walk multiple times, so we must walk it
    # only once -- however, to let us write things naturally, convert it to a
    # list and we can assume it is safe to walk repeatedly
    iterable = list(iterable)

    # extract headers and keys as separate lists
    headers = [h for (h, k) in headers_and_keys]
    keys = [k for (h, k) in headers_and_keys]

    # convert all keys to keyfuncs
    keyfuncs = [_key_to_keyfunc(key) for key in keys]

    # use the iterable to find the max width of an element for each column, in
    # the same order as the headers_and_keys array
    # use a special function to handle empty iterable
    def get_max_colwidth(kf):
        def _safelen(x):
            try:
                return len(x)
            except TypeError:
                return len(str(x))

        lengths = [_safelen(kf(i)) for i in iterable]
        if not lengths:
            return 0
        else:
            return max(lengths)

    widths = [get_max_colwidth(kf) for kf in keyfuncs]
    # handle the case in which the column header is the widest thing
    widths = [max(w, len(h)) for w, h in zip(widths, headers)]

    # create a format string based on column widths
    format_str = " | ".join("{:" + str(w) + "}" for w in widths)

    def none_to_null(val):
        if val is None:
            return "NULL"
        return val

    # print headers
    if print_headers:
        click.echo(format_str.format(*[h for h in headers]))
        click.echo(format_str.format(*["-" * w for w in widths]))
    # print the rows of data
    for i in iterable:
        click.echo(format_str.format(*[none_to_null(kf(i)) for kf in keyfuncs]))


def formatted_print(
    response_data,
    simple_text=None,
    text_preamble=None,
    text_epilog=None,
    text_format=FORMAT_TEXT_TABLE,
    json_converter=None,
    fields=None,
    response_key=None,
):
    """
    A generic output formatter. Consumes the following pieces of data:

    ``response_data`` is a dict or GlobusResponse object. It contains either an
    API response or synthesized data for printing.

    ``simple_text`` is a text override -- normal printing is skipped and this
    string is printed instead (text output only)
    ``text_preamble`` is text which prints before normal printing (text output
    only)
    ``text_epilog`` is text which prints after normal printing (text output
    only)
    ``text_format`` is one of the FORMAT_TEXT_* constants OR a callable which
    takes ``response_data`` and prints output. Note that when a callable is
    given, it does the actual printing

    ``json_converter`` is a callable that does preprocessing of JSON output. It
    must take ``response_data`` and produce another dict or dict-like object
    (json/unix output only)

    ``fields`` is an iterable of (fieldname, keyfunc) tuples. When keyfunc is
    a string, it is implicitly converted to `lambda x: x[keyfunc]` (text output
    only)

    ``response_key`` is a key into the data to print. When used with table
    printing, it must get an iterable out, and when used with raw printing, it
    gets a string. Necessary for certain formats like text table (text output
    only)
    """

    def _assert_fields():
        if fields is None:
            raise ValueError(
                "Internal Error! Output format requires fields; none given. "
                "You can workaround this error by using `--format JSON`"
            )

    def _print_as_json():
        print_json_response(
            json_converter(response_data) if json_converter else response_data
        )

    def _print_as_unix():
        print_unix_response(
            json_converter(response_data) if json_converter else response_data
        )

    def _print_as_text():
        # if we're given simple text, print that and exit
        if simple_text is not None:
            click.echo(simple_text)
            return

        # if there's a preamble, print it beofre any other text
        if text_preamble is not None:
            click.echo(text_preamble)

        # if there's a response key, key into it
        data = response_data if response_key is None else response_data[response_key]

        #  do the various kinds of printing
        if text_format == FORMAT_TEXT_TABLE:
            _assert_fields()
            print_table(data, fields)
        elif text_format == FORMAT_TEXT_RECORD:
            _assert_fields()
            colon_formatted_print(data, fields)
        elif text_format == FORMAT_TEXT_RAW:
            click.echo(data)
        elif text_format == FORMAT_TEXT_CUSTOM:
            _custom_text_formatter(data)

        # if there's an epilog, print it after any text
        if text_epilog is not None:
            click.echo(text_epilog)

    if isinstance(text_format, six.string_types):
        text_format = text_format
        _custom_text_formatter = None
    else:
        _custom_text_formatter = text_format
        text_format = FORMAT_TEXT_CUSTOM

    if outformat_is_json():
        _print_as_json()
    elif outformat_is_unix():
        _print_as_unix()
    else:
        # silent does nothing
        if text_format == FORMAT_SILENT:
            return
        _print_as_text()



import time

import globus_sdk

from globus_cli.config import AUTH_AT_OPTNAME, internal_auth_client, lookup_option
from globus_cli.parsing import command
from globus_cli.safeio import formatted_print, print_command_hint
from globus_cli.services.auth import get_auth_client


@command("show", short_help="Show your current CLI auth session")
def session_show():
    """List all identities in your current CLI auth session."""
    # get a token to introspect, refreshing if neccecary
    auth_client = internal_auth_client()
    try:
        auth_client.authorizer._check_expiration_time()
    except AttributeError:  # if we have no RefreshTokenAuthorizor
        pass
    access_token = lookup_option(AUTH_AT_OPTNAME)

    # only instance clients can introspect tokens
    if isinstance(auth_client, globus_sdk.ConfidentialAppAuthClient):
        res = auth_client.oauth2_token_introspect(access_token, include="session_info")

        session_info = res.get("session_info", {})
        authentications = session_info.get("authentications") or {}

    # empty session if still using Native App Client
    else:
        session_info = {}
        authentications = {}

    # resolve ids to human readable usernames
    resolved_ids = globus_sdk.IdentityMap(get_auth_client(), list(authentications))

    # put the nested dicts in a format table output can work with
    # while also converting vals into human readable formats
    list_data = [
        {
            "id": key,
            "username": resolved_ids.get(key, {}).get("username"),
            "auth_time": time.strftime(
                "%Y-%m-%d %H:%M %Z", time.localtime(vals["auth_time"])
            ),
        }
        for key, vals in authentications.items()
    ]

    print_command_hint(
        "For information on your primary identity or full identity set see\n"
        "  globus whoami\n"
    )

    formatted_print(
        list_data,
        json_converter=lambda x: session_info,
        fields=[("Username", "username"), ("ID", "id"), ("Auth Time", "auth_time")],
    )



import click
from globus_sdk import TransferData

from globus_cli.parsing import (
    ENDPOINT_PLUS_OPTPATH,
    TaskPath,
    command,
    shlex_process_stdin,
    task_submission_options,
)
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print
from globus_cli.services.transfer import autoactivate, get_client


@command("transfer", short_help="Submit a transfer task (asynchronous)")
@task_submission_options
@click.option(
    "--recursive",
    "-r",
    is_flag=True,
    help="SOURCE_PATH and DEST_PATH are both directories, do a recursive dir transfer",
)
@click.option(
    "--sync-level",
    "-s",
    default=None,
    show_default=True,
    type=click.Choice(("exists", "size", "mtime", "checksum"), case_sensitive=False),
    help=(
        "How will the Transfer task determine whether or not to "
        "actually transfer a file over the network?"
    ),
)
@click.option(
    "--preserve-mtime",
    is_flag=True,
    default=False,
    help="Preserve file and directory modification times.",
)
@click.option(
    "--verify-checksum/--no-verify-checksum",
    default=True,
    show_default=True,
    help="Verify checksum after transfer.",
)
@click.option(
    "--encrypt",
    is_flag=True,
    default=False,
    help="Encrypt data sent through the network.",
)
@click.option(
    "--delete",
    is_flag=True,
    default=False,
    help=(
        "Delete extraneous files in the destination directory. "
        "Only applies to recursive directory transfers."
    ),
)
@click.option(
    "--batch",
    is_flag=True,
    help=(
        "Accept a batch of source/dest path pairs on stdin (i.e. "
        "run in batchmode). "
        "Uses SOURCE_ENDPOINT_ID and DEST_ENDPOINT_ID as passed "
        "on the commandline. Commandline paths are still allowed "
        "and are used as prefixes to the batchmode inputs."
    ),
)
@click.option(
    "--external-checksum",
    help=(
        "An external checksum to verify source file and data "
        "transfer integrity. Assumed to be an MD5 checksum if "
        "--checksum-algorithm is not given."
    ),
)
@click.option(
    "--checksum-algorithm",
    default=None,
    show_default=True,
    help=("Specify an algorithm for --external-checksum or --verify-checksum"),
)
@click.argument(
    "source", metavar="SOURCE_ENDPOINT_ID[:SOURCE_PATH]", type=ENDPOINT_PLUS_OPTPATH
)
@click.argument(
    "destination", metavar="DEST_ENDPOINT_ID[:DEST_PATH]", type=ENDPOINT_PLUS_OPTPATH
)
@click.option("--perf-cc", type=int, hidden=True)
@click.option("--perf-p", type=int, hidden=True)
@click.option("--perf-pp", type=int, hidden=True)
@click.option("--perf-udt", is_flag=True, default=None, hidden=True)
def transfer_command(
    batch,
    sync_level,
    recursive,
    destination,
    source,
    checksum_algorithm,
    external_checksum,
    label,
    preserve_mtime,
    verify_checksum,
    encrypt,
    submission_id,
    dry_run,
    delete,
    deadline,
    skip_activation_check,
    notify,
    perf_cc,
    perf_p,
    perf_pp,
    perf_udt,
):
    """
    Copy a file or directory from one endpoint to another as an asynchronous
    task.

    \b
    Batched Input
    ===

    If you use `SOURCE_PATH` and `DEST_PATH` without the `--batch` flag, you
    will submit a single-file or single-directory transfer task.
    This has behavior similar to `cp` and `cp -r`, across endpoints of course.

    Using `--batch`, `globus transfer` can submit a task which transfers
    multiple files or directories. Paths to transfer are taken from stdin.
    Lines are split on spaces, respecting quotes, and every line is treated as
    a file or directory to transfer.

    \b
    Lines are of the form
    [--recursive] [--external-checksum TEXT] SOURCE_PATH DEST_PATH\n

    Skips empty lines and allows comments beginning with "#".

    \b
    If you use `--batch` and a commandline SOURCE_PATH and/or DEST_PATH, these
    paths will be used as dir prefixes to any paths on stdin.

    \b
    Sync Levels
    ===

    Sync Levels are ways to decide whether or not files are copied, with the
    following definitions:

    EXISTS: Determine whether or not to transfer based on file existence.
    If the destination file is absent, do the transfer.

    SIZE: Determine whether or not to transfer based on the size of the file.
    If destination file size does not match the source, do the transfer.

    MTIME: Determine whether or not to transfer based on modification times.
    If source has a newer modififed time than the destination, do the transfer.

    CHECKSUM: Determine whether or not to transfer based on checksums of file
    contents.
    If source and destination contents differ, as determined by a checksum of
    their contents, do the transfer.

    If a transfer fails, CHECKSUM must be used to restart the transfer.
    All other levels can lead to data corruption.
    """
    source_endpoint, cmd_source_path = source
    dest_endpoint, cmd_dest_path = destination

    if recursive and batch:
        raise click.UsageError(
            (
                "You cannot use --recursive in addition to --batch. "
                "Instead, use --recursive on lines of --batch input "
                "which need it"
            )
        )

    if external_checksum and batch:
        raise click.UsageError(
            (
                "You cannot use --external-checksum in addition to --batch. "
                "Instead, use --external-checksum on lines of --batch input "
                "which need it"
            )
        )

    if recursive and external_checksum:
        raise click.UsageError(
            "--recursive and --external-checksum are mutually exclusive"
        )

    if (cmd_source_path is None or cmd_dest_path is None) and (not batch):
        raise click.UsageError(
            "transfer requires either SOURCE_PATH and DEST_PATH or --batch"
        )

    # because python can't handle multiple **kwargs expansions in a single
    # call, we need to get a little bit clever
    # both the performance options (of which there are a few), and the
    # notification options (also there are a few) have elements which should be
    # omitted in some cases
    # notify comes to us clean, perf opts need more care
    # put them together into a dict before passing to TransferData
    kwargs = {}
    perf_opts = dict(
        (k, v)
        for (k, v) in dict(
            perf_cc=perf_cc, perf_p=perf_p, perf_pp=perf_pp, perf_udt=perf_udt
        ).items()
        if v is not None
    )
    kwargs.update(perf_opts)
    kwargs.update(notify)

    client = get_client()
    transfer_data = TransferData(
        client,
        source_endpoint,
        dest_endpoint,
        label=label,
        sync_level=sync_level,
        verify_checksum=verify_checksum,
        preserve_timestamp=preserve_mtime,
        encrypt_data=encrypt,
        submission_id=submission_id,
        delete_destination_extra=delete,
        deadline=deadline,
        skip_activation_check=skip_activation_check,
        **kwargs
    )

    if batch:

        @click.command()
        @click.option("--external-checksum")
        @click.option("--recursive", "-r", is_flag=True)
        @click.argument("source_path", type=TaskPath(base_dir=cmd_source_path))
        @click.argument("dest_path", type=TaskPath(base_dir=cmd_dest_path))
        def process_batch_line(dest_path, source_path, recursive, external_checksum):
            """
            Parse a line of batch input and turn it into a transfer submission
            item.
            """
            if recursive and external_checksum:
                raise click.UsageError(
                    "--recursive and --external-checksum are mutually exclusive"
                )
            transfer_data.add_item(
                str(source_path),
                str(dest_path),
                external_checksum=external_checksum,
                checksum_algorithm=checksum_algorithm,
                recursive=recursive,
            )

        shlex_process_stdin(
            process_batch_line,
            (
                "Enter transfers, line by line, as\n\n"
                "    [--recursive] [--external-checksum TEXT] SOURCE_PATH DEST_PATH\n"
            ),
        )
    else:
        transfer_data.add_item(
            cmd_source_path,
            cmd_dest_path,
            external_checksum=external_checksum,
            checksum_algorithm=checksum_algorithm,
            recursive=recursive,
        )

    if dry_run:
        formatted_print(
            transfer_data,
            response_key="DATA",
            fields=(
                ("Source Path", "source_path"),
                ("Dest Path", "destination_path"),
                ("Recursive", "recursive"),
                ("External Checksum", "external_checksum"),
            ),
        )
        # exit safely
        return

    # autoactivate after parsing all args and putting things together
    # skip this if skip-activation-check is given
    if not skip_activation_check:
        autoactivate(client, source_endpoint, if_expires_in=60)
        autoactivate(client, dest_endpoint, if_expires_in=60)

    res = client.submit_transfer(transfer_data)
    formatted_print(
        res,
        text_format=FORMAT_TEXT_RECORD,
        fields=(("Message", "message"), ("Task ID", "task_id")),
    )



import click
from globus_sdk.exc import AuthAPIError

from globus_cli.config import (
    AUTH_RT_OPTNAME,
    TRANSFER_RT_OPTNAME,
    internal_auth_client,
    lookup_option,
)
from globus_cli.helpers import (
    do_link_auth_flow,
    do_local_server_auth_flow,
    is_remote_session,
)
from globus_cli.parsing import command, no_local_server_option

_SHARED_EPILOG = """\

You can check your primary identity with
  globus whoami

For information on which of your identities are in session use
  globus session show

Logout of the Globus CLI with
  globus logout
"""

_LOGIN_EPILOG = (
    (
        u"""\

You have successfully logged in to the Globus CLI!
"""
    )
    + _SHARED_EPILOG
)

_LOGGED_IN_RESPONSE = (
    (
        """\
You are already logged in!

You may force a new login with
  globus login --force
"""
    )
    + _SHARED_EPILOG
)


@command(
    "login",
    short_help="Log into Globus to get credentials for the Globus CLI",
    disable_options=["format", "map_http_status"],
)
@no_local_server_option
@click.option(
    "--force",
    is_flag=True,
    help=("Do a fresh login, ignoring any existing credentials"),
)
def login_command(no_local_server, force):
    """
    Get credentials for the Globus CLI

    Necessary before any Globus CLI commands which require
    authentication will work
    """
    # if not forcing, stop if user already logged in
    if not force and check_logged_in():
        click.echo(_LOGGED_IN_RESPONSE)
        return

    # use a link login if remote session or user requested
    if no_local_server or is_remote_session():
        do_link_auth_flow(force_new_client=True)

    # otherwise default to a local server login flow
    else:
        click.echo(
            "You are running 'globus login', which should automatically open "
            "a browser window for you to login.\n"
            "If this fails or you experience difficulty, try "
            "'globus login --no-local-server'"
            "\n---"
        )
        do_local_server_auth_flow(force_new_client=True)

    # print epilog
    click.echo(_LOGIN_EPILOG)


def check_logged_in():
    # first, try to get the refresh tokens from config
    # we can skip the access tokens and their expiration times as those are not
    # strictly necessary
    transfer_rt = lookup_option(TRANSFER_RT_OPTNAME)
    auth_rt = lookup_option(AUTH_RT_OPTNAME)

    # if either of the refresh tokens are null return False
    if transfer_rt is None or auth_rt is None:
        return False

    # get or create the instance client
    auth_client = internal_auth_client(requires_instance=True)

    # check that tokens and client are valid
    try:
        for tok in (transfer_rt, auth_rt):
            res = auth_client.oauth2_validate_token(tok)
            if not res["active"]:
                return False

    # if the instance client is invalid, an AuthAPIError will be raised
    # we then force a new client to be created before continuing
    except AuthAPIError:
        return False

    return True



import os
import sys

from setuptools import find_packages, setup

if sys.version_info < (2, 7):
    raise NotImplementedError(
        """\n
##############################################################
# globus-cli does not support python versions older than 2.7 #
##############################################################"""
    )


# single source of truth for package version
version_ns = {}
with open(os.path.join("globus_cli", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns["__version__"]

setup(
    name="globus-cli",
    version=version,
    packages=find_packages(exclude=["tests", "tests.*"]),
    install_requires=[
        "globus-sdk==1.9.0",
        "click>=7.1.1,<8.0",
        "jmespath==0.9.4",
        "configobj>=5.0.6,<6.0.0",
        "requests>=2.0.0,<3.0.0",
        "six>=1.10.0,<2.0.0",
        # cryptography has unusual versioning and compatibility rules:
        # https://cryptography.io/en/latest/api-stability/
        # we trust the two next major versions, per the Deprecation policy
        #
        # as new versions of cryptography are released, we may need to update
        # this requirement
        "cryptography>=1.8.1,<3.0.0",
    ],
    extras_require={
        # deprecated, but do not remove -- doing so would break installs which
        # are already using this extra
        "delegate-proxy": [],
        # the development extra is for CLI developers only
        "development": [
            # testing
            "pytest<5",
            "pytest-cov<3.0",
            # mock on py2 and py3.5
            # not just py2: py3 versions of mock don't all have the same
            # interface!
            'mock==2.0.0;python_version<"3.6"',
        ],
    },
    entry_points={"console_scripts": ["globus = globus_cli:main"]},
    # descriptive info, non-critical
    description="Globus CLI",
    long_description=open("README.rst").read(),
    author="Stephen Rosen",
    author_email="sirosen@globus.org",
    url="https://github.com/globus/globus-cli",
    keywords=["globus", "cli", "command line"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



from globus_cli.commands.endpoint.role.create import role_create
from globus_cli.commands.endpoint.role.delete import role_delete
from globus_cli.commands.endpoint.role.list import role_list
from globus_cli.commands.endpoint.role.show import role_show
from globus_cli.parsing import group


@group("role")
def role_command():
    """Manage endpoint roles"""


role_command.add_command(role_list)
role_command.add_command(role_show)
role_command.add_command(role_create)
role_command.add_command(role_delete)



import platform
import site
import sys

import click

from globus_cli.safeio import is_verbose, verbosity
from globus_cli.version import get_versions


def _get_package_data():
    """
    Import a set of important packages and return relevant data about them in a
    dict.
    Imports are done in here to avoid potential for circular imports and other
    problems, and to make iteration simpler.
    """
    moddata = []
    modlist = (
        "click",
        "configobj",
        "cryptography",
        "globus_cli",
        "globus_sdk",
        "jmespath",
        "requests",
        "six",
    )
    if verbosity() < 2:
        modlist = ("globus_cli", "globus_sdk", "requests")

    for mod in modlist:
        cur = [mod]
        try:
            loaded_mod = __import__(mod)
        except ImportError:
            loaded_mod = None

        for attr in ("__version__", "__file__", "__path__"):
            # if loading failed, be sure to pad with error messages
            if loaded_mod is None:
                cur.append("[import failed]")
                continue

            try:
                attrval = getattr(loaded_mod, attr)
            except AttributeError:
                attrval = ""
            cur.append(attrval)
        moddata.append(cur)

    return moddata


def print_version():
    """
    Print out the current version, and at least try to fetch the latest from
    PyPi to print alongside it.

    It may seem odd that this isn't in globus_cli.version , but it's done this
    way to separate concerns over printing the version from looking it up.
    """
    latest, current = get_versions()
    if latest is None:
        click.echo(
            ("Installed Version: {0}\nFailed to lookup latest version.").format(current)
        )
    else:
        click.echo(
            ("Installed Version: {0}\nLatest Version:    {1}\n\n{2}").format(
                current,
                latest,
                "You are running the latest version of the Globus CLI"
                if current == latest
                else (
                    "You should update your version of the Globus CLI with\n"
                    "  globus update"
                )
                if current < latest
                else "You are running a preview version of the Globus CLI",
            )
        )

    # verbose shows more platform and python info
    # it also includes versions of some CLI dependencies
    if is_verbose():
        moddata = _get_package_data()

        click.echo("\nVerbose Data\n---")

        click.echo("platform:")
        click.echo("  platform: {}".format(platform.platform()))
        click.echo("  py_implementation: {}".format(platform.python_implementation()))
        click.echo("  py_version: {}".format(platform.python_version()))
        click.echo("  sys.executable: {}".format(sys.executable))
        click.echo("  site.USER_BASE: {}".format(site.USER_BASE))

        click.echo("modules:")
        for mod, modversion, modfile, modpath in moddata:
            click.echo("  {}:".format(mod))
            click.echo("    __version__: {}".format(modversion))
            click.echo("    __file__: {}".format(modfile))
            click.echo("    __path__: {}".format(modpath))



import warnings

import click
import jmespath

from globus_cli import config

# Format Enum for output formatting
# could use a namedtuple, but that's overkill
JSON_FORMAT = "json"
TEXT_FORMAT = "text"
UNIX_FORMAT = "unix"


class CommandState(object):
    def __init__(self):
        # default is config value, or TEXT if it's not set
        self.output_format = config.get_output_format() or TEXT_FORMAT
        # a jmespath expression to process on the json output
        self.jmespath_expr = None
        # default is always False
        self.debug = False
        # default is 0
        self.verbosity = 0
        # by default, empty dict
        self.http_status_map = {}

    def outformat_is_text(self):
        return self.output_format == TEXT_FORMAT

    def outformat_is_json(self):
        return self.output_format == JSON_FORMAT

    def outformat_is_unix(self):
        return self.output_format == UNIX_FORMAT

    def is_verbose(self):
        return self.verbosity > 0


def format_option(f):
    def callback(ctx, param, value):
        if not value:
            return

        state = ctx.ensure_object(CommandState)

        # when a jmespath expr is set, ignore --format=text
        if value == TEXT_FORMAT and state.jmespath_expr:
            return

        state.output_format = value.lower()

    def jmespath_callback(ctx, param, value):
        if value is None:
            return

        state = ctx.ensure_object(CommandState)
        state.jmespath_expr = jmespath.compile(value)

        if state.output_format == TEXT_FORMAT:
            state.output_format = JSON_FORMAT

    f = click.option(
        "-F",
        "--format",
        type=click.Choice(
            [UNIX_FORMAT, JSON_FORMAT, TEXT_FORMAT], case_sensitive=False
        ),
        help="Output format for stdout. Defaults to text",
        expose_value=False,
        callback=callback,
    )(f)
    f = click.option(
        "--jmespath",
        "--jq",
        help=(
            "A JMESPath expression to apply to json output. "
            "Takes precedence over any specified '--format' and forces "
            "the format to be json processed by this expression"
        ),
        expose_value=False,
        callback=jmespath_callback,
    )(f)
    return f


def debug_option(f):
    def callback(ctx, param, value):
        if not value or ctx.resilient_parsing:
            # turn off warnings altogether
            warnings.simplefilter("ignore")
            return

        warnings.simplefilter("default")
        state = ctx.ensure_object(CommandState)
        state.debug = True
        config.setup_logging(level="DEBUG")

    return click.option(
        "--debug",
        is_flag=True,
        hidden=True,
        expose_value=False,
        callback=callback,
        is_eager=True,
    )(f)


def verbose_option(f):
    def callback(ctx, param, value):
        # set state verbosity value from option
        state = ctx.ensure_object(CommandState)
        state.verbosity = value

        # no verbosity
        # all warnings are ignored
        # logging is not turned on
        if value == 0:
            warnings.simplefilter("ignore")

        # verbosity level 1
        # warnings set to once
        # logging set to error
        if value == 1:
            warnings.simplefilter("once")
            config.setup_logging(level="ERROR")

        # verbosity level 2
        # warnings set to default
        # logging set to info
        if value == 2:
            warnings.simplefilter("default")
            config.setup_logging(level="INFO")

        # verbosity level 3+
        # warnings set to always
        # logging set to debug
        # sets debug flag to true
        if value >= 3:
            warnings.simplefilter("always")
            state.debug = True
            config.setup_logging(level="DEBUG")

    return click.option(
        "--verbose",
        "-v",
        count=True,
        expose_value=False,
        callback=callback,
        is_eager=True,
        help="Control level of output",
    )(f)


def map_http_status_option(f):
    exit_stat_set = [0, 1] + list(range(50, 100))

    def per_val_callback(ctx, value):
        if value is None:
            return None
        state = ctx.ensure_object(CommandState)
        try:
            # we may be given a comma-delimited list of values
            # any cases of empty strings are dropped
            pairs = [x for x in (y.strip() for y in value.split(",")) if len(x)]
            # iterate over those pairs, splitting them on `=` signs
            for http_stat, exit_stat in (pair.split("=") for pair in pairs):
                # "parse" as ints
                http_stat, exit_stat = int(http_stat), int(exit_stat)
                # force into the desired range
                if exit_stat not in exit_stat_set:
                    raise ValueError()
                # map the status
                state.http_status_map[http_stat] = exit_stat
        # two conditions can cause ValueError: split didn't give right number
        # of args, or results weren't int()-able
        except ValueError:
            raise click.UsageError(
                "--map-http-status must have an argument of the form "
                '"INT=INT,INT=INT,..." and values of exit codes must be in '
                "0,1,50-99"
            )

    def callback(ctx, param, value):
        """
        Wrap the per-value callback -- multiple=True means that the value is
        always a tuple of given vals.
        """
        for v in value:
            per_val_callback(ctx, v)

    return click.option(
        "--map-http-status",
        help=(
            "Map HTTP statuses to any of these exit codes: 0,1,50-99. "
            'e.g. "404=50,403=51"'
        ),
        expose_value=False,
        callback=callback,
        multiple=True,
    )(f)



import logging.config
import os

import globus_sdk
from configobj import ConfigObj

__all__ = [
    # option name constants
    "OUTPUT_FORMAT_OPTNAME",
    "MYPROXY_USERNAME_OPTNAME",
    "AUTH_RT_OPTNAME",
    "AUTH_AT_OPTNAME",
    "AUTH_AT_EXPIRES_OPTNAME",
    "TRANSFER_RT_OPTNAME",
    "TRANSFER_AT_OPTNAME",
    "AUTH_AT_EXPIRES_OPTNAME",
    "CLIENT_ID_OPTNAME",
    "CLIENT_SECRET_OPTNAME",
    "GLOBUS_ENV",
    "internal_native_client",
    "internal_auth_client",
    "get_output_format",
    "get_auth_tokens",
    "get_transfer_tokens",
    "get_config_obj",
    "write_option",
    "remove_option",
    "lookup_option",
]


# constants for use whenever we need to do things using
# instance clients from the CLI Native App Template
# primarily accessed via `internal_auth_client()`
CLIENT_ID_OPTNAME = "client_id"
CLIENT_SECRET_OPTNAME = "client_secret"
TEMPLATE_ID_OPTNAME = "template_id"
DEFAULT_TEMPLATE_ID = "95fdeba8-fac2-42bd-a357-e068d82ff78e"

# constants for global use
OUTPUT_FORMAT_OPTNAME = "output_format"
MYPROXY_USERNAME_OPTNAME = "default_myproxy_username"
AUTH_RT_OPTNAME = "auth_refresh_token"
AUTH_AT_OPTNAME = "auth_access_token"
AUTH_AT_EXPIRES_OPTNAME = "auth_access_token_expires"
TRANSFER_RT_OPTNAME = "transfer_refresh_token"
TRANSFER_AT_OPTNAME = "transfer_access_token"
TRANSFER_AT_EXPIRES_OPTNAME = "transfer_access_token_expires"

# get the environment from env var (not exported)
GLOBUS_ENV = os.environ.get("GLOBUS_SDK_ENVIRONMENT")

# if the env is set, rewrite the option names to have it as a prefix
if GLOBUS_ENV:
    AUTH_RT_OPTNAME = "{0}_auth_refresh_token".format(GLOBUS_ENV)
    AUTH_AT_OPTNAME = "{0}_auth_access_token".format(GLOBUS_ENV)
    AUTH_AT_EXPIRES_OPTNAME = "{0}_auth_access_token_expires".format(GLOBUS_ENV)
    TRANSFER_RT_OPTNAME = "{0}_transfer_refresh_token".format(GLOBUS_ENV)
    TRANSFER_AT_OPTNAME = "{0}_transfer_access_token".format(GLOBUS_ENV)
    TRANSFER_AT_EXPIRES_OPTNAME = "{0}_transfer_access_token_expires".format(GLOBUS_ENV)

    CLIENT_ID_OPTNAME = "{0}_client_id".format(GLOBUS_ENV)
    CLIENT_SECRET_OPTNAME = "{0}_client_secret".format(GLOBUS_ENV)
    TEMPLATE_ID_OPTNAME = "{0}_template_id".format(GLOBUS_ENV)
    DEFAULT_TEMPLATE_ID = {
        "sandbox": "33b6a241-bce4-4359-9c6d-09f88b3c9eef",
        "integration": "e0c31fd1-663b-44e1-840f-f4304bb9ee7a",
        "test": "0ebfd058-452f-40c3-babf-5a6b16a7b337",
        "staging": "3029c3cb-c8d9-4f2b-979c-c53330aa7327",
        "preview": "b2867dbb-0846-4579-8486-dc70763d700b",
    }.get(GLOBUS_ENV, DEFAULT_TEMPLATE_ID)


def get_config_obj(system=False, file_error=False):
    if system:
        path = "/etc/globus.cfg"
    else:
        path = os.path.expanduser("~/.globus.cfg")

    conf = ConfigObj(path, encoding="utf-8", file_error=file_error)

    # delete any old whomai values in the cli section
    for key in conf.get("cli", {}):
        if "whoami_identity_" in key:
            del conf["cli"][key]
            conf.write()

    return conf


def lookup_option(option, section="cli", environment=None):
    conf = get_config_obj()
    try:
        if environment:
            return conf["environment " + environment][option]
        else:
            return conf[section][option]
    except KeyError:
        return None


def remove_option(option, section="cli", system=False):
    conf = get_config_obj(system=system)

    # if there's no section for the option we're removing, just return None
    try:
        section = conf[section]
    except KeyError:
        return None

    try:
        opt_val = section[option]

        # remove value and flush to disk
        del section[option]
        conf.write()
    except KeyError:
        opt_val = None

    # return the just-deleted value
    return opt_val


def write_option(option, value, section="cli", system=False):
    """
    Write an option to disk -- doesn't handle config reloading
    """
    # deny rwx to Group and World -- don't bother storing the returned old mask
    # value, since we'll never restore it in the CLI anyway
    # do this on every call to ensure that we're always consistent about it
    os.umask(0o077)

    # FIXME: DRY violation with config_commands.helpers
    conf = get_config_obj(system=system)

    # add the section if absent
    if section not in conf:
        conf[section] = {}

    conf[section][option] = value
    conf.write()


def get_output_format():
    return lookup_option(OUTPUT_FORMAT_OPTNAME)


def get_auth_tokens():
    expires = lookup_option(AUTH_AT_EXPIRES_OPTNAME)
    if expires is not None:
        expires = int(expires)

    return {
        "refresh_token": lookup_option(AUTH_RT_OPTNAME),
        "access_token": lookup_option(AUTH_AT_OPTNAME),
        "access_token_expires": expires,
    }


def set_auth_tokens(access_token, refresh_token, expires_at):
    write_option(AUTH_AT_OPTNAME, access_token)
    write_option(AUTH_RT_OPTNAME, refresh_token)
    write_option(AUTH_AT_EXPIRES_OPTNAME, expires_at)


def get_transfer_tokens():
    expires = lookup_option(TRANSFER_AT_EXPIRES_OPTNAME)
    if expires is not None:
        expires = int(expires)

    return {
        "refresh_token": lookup_option(TRANSFER_RT_OPTNAME),
        "access_token": lookup_option(TRANSFER_AT_OPTNAME),
        "access_token_expires": expires,
    }


def set_transfer_tokens(access_token, refresh_token, expires_at):
    write_option(TRANSFER_AT_OPTNAME, access_token)
    write_option(TRANSFER_RT_OPTNAME, refresh_token)
    write_option(TRANSFER_AT_EXPIRES_OPTNAME, expires_at)


def internal_native_client():
    template_id = lookup_option(TEMPLATE_ID_OPTNAME) or DEFAULT_TEMPLATE_ID
    return globus_sdk.NativeAppAuthClient(template_id)


def internal_auth_client(requires_instance=False, force_new_client=False):
    """
    Looks up the values for this CLI's Instance Client in config

    If none exists and requires_instance is True or force_new_client is True,
    registers a new Instance Client with Globus Auth

    If none exists and requires_instance is false, defaults to a Native Client
    for backwards compatibility

    Returns either a NativeAppAuthClient or a ConfidentialAppAuthClient
    """
    client_id = lookup_option(CLIENT_ID_OPTNAME)
    client_secret = lookup_option(CLIENT_SECRET_OPTNAME)
    template_id = lookup_option(TEMPLATE_ID_OPTNAME) or DEFAULT_TEMPLATE_ID
    template_client = internal_native_client()
    existing = client_id and client_secret

    # if we are forcing a new client, delete any existing client
    if force_new_client and existing:
        existing_client = globus_sdk.ConfidentialAppAuthClient(client_id, client_secret)
        try:
            existing_client.delete("/v2/api/clients/{}".format(client_id))

        # if the client secret has been invalidated or the client has
        # already been removed, we continue on
        except globus_sdk.exc.AuthAPIError:
            pass

    # if we require a new client to be made
    if force_new_client or (requires_instance and not existing):
        # register a new instance client with auth
        body = {"client": {"template_id": template_id, "name": "Globus CLI"}}
        res = template_client.post("/v2/api/clients", json_body=body)

        # get values and write to config
        credential_data = res["included"]["client_credential"]
        client_id = credential_data["client"]
        client_secret = credential_data["secret"]
        write_option(CLIENT_ID_OPTNAME, client_id)
        write_option(CLIENT_SECRET_OPTNAME, client_secret)

        return globus_sdk.ConfidentialAppAuthClient(
            client_id, client_secret, app_name="Globus CLI"
        )

    # if we already have a client, just return it
    elif existing:
        return globus_sdk.ConfidentialAppAuthClient(
            client_id, client_secret, app_name="Globus CLI"
        )

    # fall-back to a native client to not break old logins
    # TOOD: eventually remove this behavior
    else:
        return template_client


def setup_logging(level="DEBUG"):
    conf = {
        "version": 1,
        "formatters": {
            "basic": {"format": "[%(levelname)s] %(name)s::%(funcName)s() %(message)s"}
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": level,
                "formatter": "basic",
            }
        },
        "loggers": {"globus_sdk": {"level": level, "handlers": ["console"]}},
    }

    logging.config.dictConfig(conf)



import click

from globus_cli.commands.bookmark.helpers import resolve_id_or_name
from globus_cli.parsing import command
from globus_cli.safeio import formatted_print
from globus_cli.services.transfer import get_client


@command("rename")
@click.argument("bookmark_id_or_name")
@click.argument("new_bookmark_name")
def bookmark_rename(bookmark_id_or_name, new_bookmark_name):
    """Change a bookmark's name"""
    client = get_client()
    bookmark_id = resolve_id_or_name(client, bookmark_id_or_name)["id"]

    submit_data = {"name": new_bookmark_name}

    res = client.update_bookmark(bookmark_id, submit_data)
    formatted_print(res, simple_text="Success")



import click

from globus_cli.parsing import command, endpoint_id_arg
from globus_cli.safeio import FORMAT_TEXT_RAW, formatted_print
from globus_cli.services.transfer import get_client


@command("delete")
@endpoint_id_arg
@click.argument("rule_id")
def delete_command(endpoint_id, rule_id):
    """Delete an access control rule, removing permissions"""
    client = get_client()

    res = client.delete_endpoint_acl_rule(endpoint_id, rule_id)
    formatted_print(res, text_format=FORMAT_TEXT_RAW, response_key="message")



import re

from globus_sdk import AuthClient, RefreshTokenAuthorizer

from globus_cli import version
from globus_cli.config import get_auth_tokens, internal_auth_client, set_auth_tokens

# what qualifies as a valid Identity Name?
_IDENTITY_NAME_REGEX = r"^[a-zA-Z0-9]+.*@[a-zA-z0-9-]+\..*[a-zA-Z]+$"


def _update_tokens(token_response):
    tokens = token_response.by_resource_server["auth.globus.org"]
    set_auth_tokens(
        tokens["access_token"], tokens["refresh_token"], tokens["expires_at_seconds"]
    )


def is_valid_identity_name(identity_name):
    """
    Check if a string is a valid identity name.
    Does not do any preprocessing of the identity name, so you must do so
    before invocation.
    """
    if re.match(_IDENTITY_NAME_REGEX, identity_name) is None:
        return False
    else:
        return True


def get_auth_client():
    tokens = get_auth_tokens()
    authorizer = None

    # if there's a refresh token, use it to build the authorizer
    if tokens["refresh_token"] is not None:
        authorizer = RefreshTokenAuthorizer(
            tokens["refresh_token"],
            internal_auth_client(),
            tokens["access_token"],
            tokens["access_token_expires"],
            on_refresh=_update_tokens,
        )

    client = AuthClient(authorizer=authorizer, app_name=version.app_name)
    return client


def _lookup_identity_field(id_name=None, id_id=None, field="id", provision=False):
    assert (id_name or id_id) and not (id_name and id_id)
    client = get_auth_client()

    kw = dict(provision=provision)
    if id_name:
        kw.update({"usernames": id_name})
    else:
        kw.update({"ids": id_id})

    try:
        return client.get_identities(**kw)["identities"][0][field]
    except (IndexError, KeyError):
        # IndexError: identity does not exist and wasn't provisioned
        # KeyError: `field` does not exist for the requested identity
        return None


def maybe_lookup_identity_id(identity_name, provision=False):
    if is_valid_identity_name(identity_name):
        return _lookup_identity_field(id_name=identity_name, provision=provision)
    else:
        return identity_name


def lookup_identity_name(identity_id):
    return _lookup_identity_field(id_id=identity_id, field="username")



from globus_sdk import IdentityMap

from globus_cli.parsing import command, endpoint_id_arg
from globus_cli.safeio import formatted_print
from globus_cli.services.auth import get_auth_client
from globus_cli.services.transfer import get_client


@command("list")
@endpoint_id_arg
def list_command(endpoint_id):
    """List of permissions on an endpoint"""
    client = get_client()

    rules = client.endpoint_acl_list(endpoint_id)

    resolved_ids = IdentityMap(
        get_auth_client(),
        (x["principal"] for x in rules if x["principal_type"] == "identity"),
    )

    def principal_str(rule):
        principal = rule["principal"]
        if rule["principal_type"] == "identity":
            try:
                return resolved_ids[principal]["username"]
            except KeyError:
                return principal
        if rule["principal_type"] == "group":
            return (u"https://app.globus.org/groups/{}").format(principal)
        return rule["principal_type"]

    formatted_print(
        rules,
        fields=[
            ("Rule ID", "id"),
            ("Permissions", "permissions"),
            ("Shared With", principal_str),
            ("Path", "path"),
        ],
    )



from globus_cli.parsing import (
    command,
    endpoint_create_and_update_params,
    endpoint_id_arg,
    validate_endpoint_create_and_update_params,
)
from globus_cli.safeio import FORMAT_TEXT_RAW, formatted_print
from globus_cli.services.transfer import assemble_generic_doc, get_client


@command("update")
@endpoint_id_arg
@endpoint_create_and_update_params(create=False)
def endpoint_update(**kwargs):
    """Update attributes of an endpoint"""
    # validate params. Requires a get call to check the endpoint type
    client = get_client()
    endpoint_id = kwargs.pop("endpoint_id")
    get_res = client.get_endpoint(endpoint_id)

    if get_res["host_endpoint_id"]:
        endpoint_type = "shared"
    elif get_res["is_globus_connect"]:
        endpoint_type = "personal"
    elif get_res["s3_url"]:
        endpoint_type = "s3"
    else:
        endpoint_type = "server"
    validate_endpoint_create_and_update_params(
        endpoint_type, get_res["subscription_id"], kwargs
    )

    # make the update
    ep_doc = assemble_generic_doc("endpoint", **kwargs)
    res = client.update_endpoint(endpoint_id, ep_doc)
    formatted_print(res, text_format=FORMAT_TEXT_RAW, response_key="message")



from globus_cli.commands.endpoint.permission.create import create_command
from globus_cli.commands.endpoint.permission.delete import delete_command
from globus_cli.commands.endpoint.permission.list import list_command
from globus_cli.commands.endpoint.permission.show import show_command
from globus_cli.commands.endpoint.permission.update import update_command
from globus_cli.parsing import group


@group("permission")
def permission_command():
    """Manage endpoint permissions (Access Control Lists)"""


permission_command.add_command(list_command)
permission_command.add_command(create_command)
permission_command.add_command(show_command)
permission_command.add_command(update_command)
permission_command.add_command(delete_command)



import click

from globus_cli.parsing import command, task_id_arg
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print
from globus_cli.services.transfer import get_client

EXPLICIT_PAUSE_MSG_FIELDS = [
    ("Source Endpoint", "source_pause_message"),
    ("Source Shared Endpoint", "source_pause_message_share"),
    ("Destination Endpoint", "destination_pause_message"),
    ("Destination Shared Endpoint", "destination_pause_message_share"),
]

PAUSE_RULE_OPERATION_FIELDS = [
    ("write", "pause_task_transfer_write"),
    ("read", "pause_task_transfer_read"),
    ("delete", "pause_task_delete"),
    ("rename", "pause_rename"),
    ("mkdir", "pause_mkdir"),
    ("ls", "pause_ls"),
]

PAUSE_RULE_DISPLAY_FIELDS = [
    (
        "Operations",
        lambda rule: "/".join(
            label for label, key in PAUSE_RULE_OPERATION_FIELDS if rule[key]
        ),
    ),
    ("On Endpoint", "endpoint_display_name"),
    ("All Users", lambda rule: "No" if rule["identity_id"] else "Yes"),
    ("Message", "message"),
]


@command("pause-info", short_help="Show why an in-progress task is currently paused")
@task_id_arg
def task_pause_info(task_id):
    """
    Show messages from activity managers who have explicitly paused the given
    in-progress task and list any active pause rules that apply to it
    """
    client = get_client()
    res = client.task_pause_info(task_id)

    def _custom_text_format(res):
        explicit_pauses = [
            field
            for field in EXPLICIT_PAUSE_MSG_FIELDS
            # n.b. some keys are absent for completed tasks
            if res.get(field[1])
        ]
        effective_pause_rules = res["pause_rules"]

        if not explicit_pauses and not effective_pause_rules:
            click.echo("Task {} is not paused.".format(task_id))
            click.get_current_context().exit(0)

        if explicit_pauses:
            formatted_print(
                res,
                fields=explicit_pauses,
                text_format=FORMAT_TEXT_RECORD,
                text_preamble="This task has been explicitly paused.\n",
                text_epilog="\n" if effective_pause_rules else None,
            )

        if effective_pause_rules:
            formatted_print(
                effective_pause_rules,
                fields=PAUSE_RULE_DISPLAY_FIELDS,
                text_preamble=(
                    "The following pause rules are effective on this task:\n"
                ),
            )

    formatted_print(res, text_format=_custom_text_format)



import click

from globus_cli.parsing import command, task_id_arg
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print
from globus_cli.services.transfer import get_client, iterable_response_to_dict

COMMON_FIELDS = [
    ("Label", "label"),
    ("Task ID", "task_id"),
    ("Is Paused", "is_paused"),
    ("Type", "type"),
    ("Directories", "directories"),
    ("Files", "files"),
    ("Status", "status"),
    ("Request Time", "request_time"),
    ("Faults", "faults"),
    ("Total Subtasks", "subtasks_total"),
    ("Subtasks Succeeded", "subtasks_succeeded"),
    ("Subtasks Pending", "subtasks_pending"),
    ("Subtasks Retrying", "subtasks_retrying"),
    ("Subtasks Failed", "subtasks_failed"),
    ("Subtasks Canceled", "subtasks_canceled"),
    ("Subtasks Expired", "subtasks_expired"),
]

ACTIVE_FIELDS = [("Deadline", "deadline"), ("Details", "nice_status")]

COMPLETED_FIELDS = [("Completion Time", "completion_time")]

DELETE_FIELDS = [
    ("Endpoint", "source_endpoint_display_name"),
    ("Endpoint ID", "source_endpoint_id"),
]

TRANSFER_FIELDS = [
    ("Source Endpoint", "source_endpoint_display_name"),
    ("Source Endpoint ID", "source_endpoint_id"),
    ("Destination Endpoint", "destination_endpoint_display_name"),
    ("Destination Endpoint ID", "destination_endpoint_id"),
    ("Bytes Transferred", "bytes_transferred"),
    ("Bytes Per Second", "effective_bytes_per_second"),
]

SUCCESSFULL_TRANSFER_FIELDS = [
    ("Source Path", "source_path"),
    ("Destination Path", "destination_path"),
]


def print_successful_transfers(client, task_id):
    res = client.task_successful_transfers(task_id, num_results=None)
    formatted_print(
        res,
        fields=SUCCESSFULL_TRANSFER_FIELDS,
        json_converter=iterable_response_to_dict,
    )


def print_task_detail(client, task_id):
    res = client.get_task(task_id)
    formatted_print(
        res,
        text_format=FORMAT_TEXT_RECORD,
        fields=(
            COMMON_FIELDS
            + (COMPLETED_FIELDS if res["completion_time"] else ACTIVE_FIELDS)
            + (DELETE_FIELDS if res["type"] == "DELETE" else TRANSFER_FIELDS)
        ),
    )


@command("show")
@task_id_arg
@click.option(
    "--successful-transfers",
    "-t",
    is_flag=True,
    default=False,
    help="Show files that were transferred as result of this task.",
)
def show_task(successful_transfers, task_id):
    """Show detailed information about a specific task"""
    client = get_client()

    if successful_transfers:
        print_successful_transfers(client, task_id)
    else:
        print_task_detail(client, task_id)



import copy

import click

_command_length = 16


@click.command(
    "list-commands",
    short_help="List all CLI Commands",
    help=(
        "List all Globus CLI Commands with short help output. "
        "For full command help, run the command with the "
        "`--help` flag"
    ),
)
def list_commands():
    def _print_cmd(command):
        # print commands with short_help
        indent = 4
        min_space = 2

        short_help = command.get_short_help_str()

        # if the output would be pinched too close together, or if the command
        # name would overflow, use two separate lines
        if len(command.name) > _command_length - min_space:
            click.echo(" " * indent + command.name)
            click.echo(" " * (indent + _command_length) + short_help)
        # otherwise, it's all cool to cram into one line, just ljust command
        # names so that they form a nice column
        else:
            click.echo(
                " " * indent
                + "{}{}".format(command.name.ljust(_command_length), short_help)
            )

    def _print_cmd_group(command, parent_names):
        parents = " ".join(parent_names)
        if parents:
            parents = parents + " "
        click.echo("\n=== {}{} ===\n".format(parents, command.name))

    def _recursive_list_commands(command, parent_names=None):
        if parent_names is None:
            parent_names = []

        # names of parent commands, including this one, for passthrough to
        # recursive calls
        new_parent_names = copy.copy(parent_names) + [command.name]

        # group-style commands are printed as headers
        if isinstance(command, click.MultiCommand):
            _print_cmd_group(command, parent_names)

            # get the set of subcommands and recursively print all of them
            group_cmds = [
                v
                for v in command.commands.values()
                if isinstance(v, click.MultiCommand)
            ]
            func_cmds = [v for v in command.commands.values() if v not in group_cmds]
            # we want to print them all, but func commands first
            for cmd in func_cmds + group_cmds:
                _recursive_list_commands(cmd, parent_names=new_parent_names)

        # individual commands are printed solo
        else:
            _print_cmd(command)

    # get the root context (the click context for the entire CLI tree)
    root_ctx = click.get_current_context().find_root()

    _recursive_list_commands(root_ctx.command)
    # get an extra newline at the end
    click.echo("")



import json

import click

from globus_cli.parsing import command, task_id_arg
from globus_cli.safeio import formatted_print
from globus_cli.services.transfer import get_client, iterable_response_to_dict


@command("event-list")
@task_id_arg
@click.option("--limit", default=10, show_default=True, help="Limit number of results.")
@click.option("--filter-errors", is_flag=True, help="Filter results to errors")
@click.option("--filter-non-errors", is_flag=True, help="Filter results to non errors")
def task_event_list(task_id, limit, filter_errors, filter_non_errors):
    """List Events for a given task"""
    client = get_client()

    # cannot filter by both errors and non errors
    if filter_errors and filter_non_errors:
        raise click.UsageError("Cannot filter by both errors and non errors")

    elif filter_errors:
        filter_string = "is_error:1"

    elif filter_non_errors:
        filter_string = "is_error:0"

    else:
        filter_string = ""

    event_iterator = client.task_event_list(
        task_id, num_results=limit, filter=filter_string
    )

    def squashed_json_details(x):
        is_json = False
        try:
            loaded = json.loads(x["details"])
            is_json = True
        except ValueError:
            loaded = x["details"]

        if is_json:
            return json.dumps(loaded, separators=(",", ":"), sort_keys=True)
        else:
            return loaded.replace("\n", "\\n")

    formatted_print(
        event_iterator,
        fields=(
            ("Time", "time"),
            ("Code", "code"),
            ("Is Error", "is_error"),
            ("Details", squashed_json_details),
        ),
        json_converter=iterable_response_to_dict,
    )



import click


def _normpath(path):
    """
    Globus Transfer-specific normalization, based on a careful reading of the stdlib
    posixpath implementation:
      https://github.com/python/cpython/blob/ea0f7aa47c5d2e58dc99314508172f0523e144c6/Lib/posixpath.py#L338

    this must be done without using os.path.normpath to be compatible with CLI calls
    from Windows systems

    Transfer requires forward slashes, even when communicating with Windows systems, so
    we must handle these strings appropriately.
    Note that this does not preserve leading slashes in the same way as python's
    posixpath module -- it's not clear how Transfer would treat such paths and
    non-obvious that we need to allow such usage

    Also, unlike normpath, we want to preserve trailing slashes because they may be
    required
    """
    initial_slash = 1 if path.startswith("/") else 0
    trailing_slash = 1 if path.endswith("/") and path != "/" else 0
    parts = path.split("/")
    new_parts = []
    for part in parts:
        if part in ("", "."):
            continue
        # either not adding a ".." OR chaining together multiple ".."s
        # OR working with a non-absolute path that starts with ".."
        if (
            part != ".."
            or (new_parts and new_parts[-1] == "..")
            or (not initial_slash and not new_parts)
        ):
            new_parts.append(part)
        elif new_parts:  # adding a ".." to a path which isn't already ending in one
            new_parts.pop()

    return ("/" * initial_slash) + "/".join(new_parts) + ("/" * trailing_slash)


def _pathjoin(a, b):
    """
    POSIX-like path join for Globus Transfer paths

    As with _normpath above, this is meant to behave correctly even on Windows systems
    """
    if not b:  # given "" as a file path
        return a
    elif b.startswith("/"):  # a path starting with / is absolute
        return b

    if a.endswith("/"):
        return a + b
    else:
        return a + "/" + b


class TaskPath(click.ParamType):
    def __init__(
        self, base_dir=None, coerce_to_dir=False, normalize=True, require_absolute=False
    ):
        """
        Task Paths are paths for passing into Transfer or Delete tasks.
        They're only slightly more than string types: they can join themselves
        with a base dir path, and they can coerce themselves to the dir format
        by appending a trailing slash if it's absent.

        For us to toggle and talk about behaviors as necessary, normalization
        is an option that defaults to True.
        Also can enforce that the path is absolute.
        """
        self.base_dir = base_dir
        self.coerce_to_dir = coerce_to_dir
        self.normalize = normalize
        self.require_absolute = require_absolute

        # the "real value" of this path holder
        self.path = None
        # the original path, as consumed before processing
        self.orig_path = None

    def convert(self, value, param, ctx):
        if value is None or (ctx and ctx.resilient_parsing):
            return
        if isinstance(value, TaskPath):
            return value

        self.orig_path = self.path = value

        if self.base_dir:
            self.path = _pathjoin(self.base_dir, self.path)
        if self.coerce_to_dir and not self.path.endswith("/"):
            self.path += "/"
        if self.normalize:
            self.path = _normpath(self.path)

        if self.require_absolute and not (
            self.path.startswith("/") or self.path.startswith("~")
        ):
            self.fail(
                "{} is not absolute (abspath required)".format(self.path),
                param=param,
                ctx=ctx,
            )

        return self

    def __repr__(self):
        return "TaskPath({})".format(
            ",".join(
                "{}={}".format(name, getattr(self, name))
                for name in (
                    "base_dir",
                    "coerce_to_dir",
                    "normalize",
                    "path",
                    "orig_path",
                )
            )
        )

    def __str__(self):
        return str(self.path)



from distutils.version import LooseVersion

# single source of truth for package version,
# see https://packaging.python.org/en/latest/single_source_version/
__version__ = "1.12.0"

# app name to send as part of SDK requests
app_name = "Globus CLI v{}".format(__version__)


# pull down version data from PyPi
def get_versions():
    """
    Wrap in a function to ensure that we don't run this every time a CLI
    command runs (yuck!)

    Also protects import of `requests` from issues when grabbed by setuptools.
    More on that inline
    """
    # import in the func (rather than top-level scope) so that at setup time,
    # `requests` isn't required -- otherwise, setuptools will fail to run
    # because requests isn't installed yet.
    import requests

    try:
        version_data = requests.get(
            "https://pypi.python.org/pypi/globus-cli/json"
        ).json()
        latest = max(LooseVersion(v) for v in version_data["releases"])
        return latest, LooseVersion(__version__)
    # if the fetch from pypi fails
    except requests.RequestException:
        return None, LooseVersion(__version__)



from textwrap import dedent

import click

from globus_cli.parsing import command, endpoint_id_arg
from globus_cli.safeio import FORMAT_TEXT_RAW, formatted_print
from globus_cli.services.transfer import get_client, get_endpoint_w_server_list


def _spec_to_matches(server_list, server_spec, mode):
    """
    mode is in {uri, hostname, hostname_port}

    A list of matching server docs.
    Should usually be 0 or 1 matches. Multiple matches are possible though.
    """
    assert mode in ("uri", "hostname", "hostname_port")

    def match(server_doc):
        if mode == "hostname":
            return server_spec == server_doc["hostname"]
        elif mode == "hostname_port":
            return server_spec == "{}:{}".format(
                server_doc["hostname"], server_doc["port"]
            )
        elif mode == "uri":
            return server_spec == "{}://{}:{}".format(
                server_doc["scheme"], server_doc["hostname"], server_doc["port"]
            )
        else:
            raise NotImplementedError("Unreachable error! Something is very wrong.")

    return [server_doc for server_doc in server_list if match(server_doc)]


def _detect_mode(server):
    try:
        int(server)
        return "id"
    except ValueError:
        pass

    if "://" in server:
        return "uri"

    if ":" in server:
        return "hostname_port"

    return "hostname"


@command("delete")
@endpoint_id_arg
@click.argument("server")
def server_delete(endpoint_id, server):
    """
    Delete a server belonging to an endpoint

    SERVER may be a server ID, HOSTNAME, HOSTNAME:PORT, or URI
    ( SCHEME://HOSTNAME:PORT )
    """
    client = get_client()

    mode = _detect_mode(server)

    # list (even if not necessary) in order to make errors more consistent when
    # mode='id'
    endpoint, server_list = get_endpoint_w_server_list(endpoint_id)

    if server_list == "S3":
        raise click.UsageError("You cannot delete servers from S3 endpoints.")

    # we don't *have to* raise an error in the GCP case, since the API would
    # deny it too, but doing so makes our errors a little bit more consistent
    # with deletes against S3 endpoints and shares
    if endpoint["is_globus_connect"]:
        raise click.UsageError(
            "You cannot delete servers from Globus Connect Personal endpoints"
        )

    if mode != "id":
        matches = _spec_to_matches(server_list, server, mode)
        if not matches:
            raise click.UsageError('No server was found matching "{}"'.format(server))
        elif len(matches) > 1:
            raise click.UsageError(
                dedent(
                    """\
                Multiple servers matched "{}":
                    {}
            """
                ).format(server, [x["id"] for x in matches])
            )
        else:
            server = matches[0]["id"]

    response = client.delete_endpoint_server(endpoint_id, server)

    formatted_print(response, text_format=FORMAT_TEXT_RAW, response_key="message")



import click

from globus_cli.parsing import command, endpoint_id_arg
from globus_cli.safeio import FORMAT_TEXT_RECORD, formatted_print
from globus_cli.services.auth import lookup_identity_name
from globus_cli.services.transfer import get_client


def _shared_with_keyfunc(rule):
    if rule["principal_type"] == "identity":
        return lookup_identity_name(rule["principal"])
    elif rule["principal_type"] == "group":
        return u"https://app.globus.org/groups/{}".format(rule["principal"])
    else:
        return rule["principal_type"]


@command("show")
@endpoint_id_arg
@click.argument("rule_id")
def show_command(endpoint_id, rule_id):
    """Show a permission on an endpoint"""
    client = get_client()

    rule = client.get_endpoint_acl_rule(endpoint_id, rule_id)
    formatted_print(
        rule,
        text_format=FORMAT_TEXT_RECORD,
        fields=(
            ("Rule ID", "id"),
            ("Permissions", "permissions"),
            ("Shared With", _shared_with_keyfunc),
            ("Path", "path"),
        ),
    )



from globus_cli.commands.task.cancel import cancel_task
from globus_cli.commands.task.event_list import task_event_list
from globus_cli.commands.task.generate_submission_id import generate_submission_id
from globus_cli.commands.task.list import task_list
from globus_cli.commands.task.pause_info import task_pause_info
from globus_cli.commands.task.show import show_task
from globus_cli.commands.task.update import update_task
from globus_cli.commands.task.wait import task_wait
from globus_cli.parsing import group


@group("task")
def task_command():
    """Manage asynchronous tasks"""


task_command.add_command(task_list)
task_command.add_command(show_task)
task_command.add_command(update_task)
task_command.add_command(cancel_task)
task_command.add_command(task_event_list)
task_command.add_command(task_pause_info)
task_command.add_command(task_wait)
task_command.add_command(generate_submission_id)
